#include <cmath>
#include <string>
#include <vector>
#include <deque>
#include <list>
#include <map>

#include <unistd.h>   // mknod, unlink, write
#include <stdio.h>
#include <sys/stat.h> // S_IFIFO
#include <fcntl.h>    // fcntl
#include <stdlib.h>   // setenv
#include <string.h>   // strrchr
#include <sys/file.h> // flock
#include <cstdlib>    // free()
#include <errno.h>

#ifndef __WIN32__
# include <sys/poll.h> // poll
#endif

#include <stdint.h>

/* Note: This module assumes everyone uses BGR16 as display depth */

static std::string VIDEO_CMD = "";
/*
-rawvideo on:fps=60:format=0x42475220:w=256:h=224:size=$[1024*224]
-audiofile "+AUDIO_FN+"
*/
static std::string AUDIO_FN = "s.log";

static bool Terminate=false;
static unsigned videonumber = 0;

unsigned long long A_sent = 0;
unsigned long long V_sent = 0;

static unsigned NonblockWrite(FILE* fp, const unsigned char*buf, unsigned length)
{
  Retry:
    int result = write(fileno(fp), buf, length);
    if(result == -1 && errno==EAGAIN)
    {
        return 0;
    }
    if(result == -1 && errno==EINTR) goto Retry;
    if(result == -1)
    {
        perror("write");
        Terminate=true;
        return 0;
    }
    return result;
}
static int PollTimeout = 2000;
static int WaitUntilOneIsWritable(FILE*f1, FILE*f2, int whichmask = 3)
{
#ifndef __WIN32__
    struct pollfd po[2] = { {fileno(f1),POLLOUT,0}, {fileno(f2),POLLOUT,0} };

    pollfd* po_ptr = po;
    unsigned po_n = 2;

    if(whichmask == 1) // f1 only
        { po_n = 1; po[1].revents = 0; }
    else if(whichmask == 2) // f2 only
        { po_ptr += 1; po[0].revents = 0; }

    poll(po_ptr, po_n, PollTimeout);

    return ((po[0].revents & POLLOUT) ? 1 : 0)
         | ((po[1].revents & POLLOUT) ? 2 : 0);
#else
    return 3;
#endif
}

static const char* USE_IMGFMT = "bgr16";
static unsigned INPUT_BPP  = 16;

#define u32(n) (n)&255,((n)>>8)&255,((n)>>16)&255,((n)>>24)&255
#define u16(n) (n)&255,((n)>>8)&255
#define s4(s) s[0],s[1],s[2],s[3]

static const unsigned FPS_SCALE = 0x1000000;

static struct Construct
{
    Construct()
    {
        char Buf[4096];
        getcwd(Buf,sizeof(Buf));
        Buf[sizeof(Buf)-1]=0;
        AUDIO_FN = Buf + std::string("/") + AUDIO_FN;
    }
} Construct;

class AVI
{
public:
    AVI()          { }
    virtual ~AVI() { }

    virtual void AudioPtr
        (unsigned r,unsigned b,unsigned c,
         const unsigned char*d, unsigned nsamples) = 0;

    virtual void VideoPtr
        (unsigned w,unsigned h,unsigned f, const unsigned char*d) = 0;

    virtual void SaveState(const std::string&) { }
    virtual void LoadState(const std::string&) { }
};

class NormalAVI: public AVI
{
    FILE* vidfp;
    FILE* audfp;

    bool KnowVideo;
    unsigned vid_width;
    unsigned vid_height;
    unsigned vid_fps_scaled;
    struct DataRec
    {
        size_t size;
        size_t done;
        const unsigned char* data;
    };

    std::list< DataRec > VideoBuffer;
    unsigned VidBufSize;

    bool KnowAudio;
    unsigned aud_rate;
    unsigned aud_chans;
    unsigned aud_bits;
    std::list< DataRec > AudioBuffer;
    unsigned AudBufSize;

public:
    NormalAVI() :
        vidfp(NULL),
        audfp(NULL),
        KnowVideo(false), VidBufSize(0),
        KnowAudio(false), AudBufSize(0)
    {
    }
    virtual ~NormalAVI()
    {
        while(VidBufSize && AudBufSize)
        {
            CheckFlushing();
        }
        if(audfp) fclose(audfp);
        if(vidfp) pclose(vidfp);
        unlink(AUDIO_FN.c_str());
    }

    virtual void AudioPtr
        (unsigned r,unsigned b,unsigned c,
         const unsigned char*d, unsigned nsamples)
    {
        if(Terminate) return;
        if(!KnowAudio)
        {
            aud_rate = r;
            aud_chans = c;
            aud_bits = b;
            KnowAudio = true;
        }
        CheckFlushing();

        unsigned bytes = nsamples * aud_chans * (aud_bits / 8);

        unsigned wrote = 0;
        if(KnowVideo && AudioBuffer.empty())
        {
            //fprintf(stderr, "Writing %u of %s from %p to %p\t", bytes, "aud", (void*)d, (void*)audfp);
            wrote = NonblockWrite(audfp, d, bytes);
            //fprintf(stderr, "Wrote %u\n", wrote);
            A_sent += wrote;
        }
        if(wrote < bytes)
        {
            //fprintf(stderr, "Buffering %u of %s (%p..%p)\n", bytes-wrote, "aud", d+wrote, d+bytes);
            DataRec rec;
            rec.size = bytes;
            rec.done = wrote;
            rec.data = d;
            AudioBuffer.push_back(rec);
            AudBufSize += bytes - wrote;
        }
        else
            std::free( (void*) d);
        CheckFlushing();
    }

    virtual void VideoPtr
        (unsigned w,unsigned h,unsigned f, const unsigned char*d)
    {
        if(Terminate) return;
        if(!KnowVideo)
        {
            vid_width      = w;
            vid_height     = h;
            vid_fps_scaled = f;
            KnowVideo = true;
        }
        CheckFlushing();

        unsigned bpp   = INPUT_BPP; if(bpp == 15 || bpp == 17) bpp = 16;
        unsigned bytes = vid_width * vid_height * bpp / 8;

        //std::vector<unsigned char> tmp(bytes, 'k');
        //d = &tmp[0];

        unsigned wrote = 0;
        if(KnowAudio && VideoBuffer.empty())
        {
            CheckBegin();
            //fprintf(stderr, "Writing %u of %s from %p to %p\t", bytes, "vid", (void*)d, (void*)vidfp);
            wrote = NonblockWrite(vidfp, d, bytes);
            //fprintf(stderr, "Wrote %u\n", wrote);
            V_sent += wrote;
        }

        if(wrote < bytes)
        {
            //fprintf(stderr, "Buffering %u of %s (%p..%p)\n", bytes-wrote, "vid", d+wrote, d+bytes);
            DataRec rec;
            rec.size = bytes;
            rec.done = wrote;
            rec.data = d;
            VideoBuffer.push_back(rec);
            VidBufSize += bytes - wrote;
        }
        CheckFlushing();
    }

private:
    /* fp is passed as a reference because it may be NULL
     * prior to calling, and this function changes it. */
    template<typename BufType>
    bool FlushBufferSome(BufType& List, unsigned& Size, FILE*& fp, const char* what)
    {
        what=what;

    Retry:
        if(List.empty() || Terminate) return false;

        /*if(List.begin()->done >= List.begin()->size)
        {
            std::free( List.begin()->data );
            List.erase(List.begin());
            goto Retry;
        }*/

        CheckBegin();

        typename BufType::iterator i = List.begin();
        DataRec& rec = *i;

        size_t bytes = rec.size - rec.done;

        //fprintf(stderr, "Writing %u of %s from %p to %p\t", bytes, what, (void*)&buf[0], (void*)fp);

        unsigned ate = NonblockWrite(fp, &rec.data[rec.done], bytes);
        if(ate == 0)
            return false;

        //fprintf(stderr, "Wrote %u\n", ate);

        if(what[0] == 'v')
            V_sent += ate;
        else
            A_sent += ate;

        rec.done += ate;
        Size -= ate;

        if(rec.done >= rec.size)
        {
            std::free( (void*) rec.data );
            List.erase(i);
        }

        return true;
    }

    void CheckFlushing()
    {
        //AudioBuffer.clear();
        //VideoBuffer.clear();

        if(KnowAudio && KnowVideo && !Terminate)
        {
            unsigned nloops = 0;
            while ((!AudioBuffer.empty() && !VideoBuffer.empty())
                || (VideoBuffer.size() >= 50
                 && nloops++ < 10))
            {
                /* vidfp = &1, audfp = &2 */
                int attempt = WaitUntilOneIsWritable(vidfp, audfp,
                    (AudioBuffer.empty() ? 0 : 2)
                  | (VideoBuffer.empty() ? 0 : 1)
                   );

                if(attempt <= 0)
                {
                    if(PollTimeout > 250) PollTimeout /= 2;
                    break; /* Some kind of error can cause this */
                }

                // Flush Audio
                if(attempt&2) FlushBufferSome(AudioBuffer, AudBufSize, audfp, "aud");

                // Flush Video
                if(attempt&1) FlushBufferSome(VideoBuffer, VidBufSize, vidfp, "vid");
                if(PollTimeout < 5000)
                    PollTimeout *= 2;
            }

            while(FlushBufferSome(VideoBuffer, VidBufSize, vidfp, "vid")) {}
            while(FlushBufferSome(AudioBuffer, AudBufSize, audfp, "aud")) {}

            /**/
            fprintf(stderr, "Buffer Sizes: Audio %u(%u) video %u(%u) -- sent A=%llu, V=%llu\n",
                (unsigned)AudioBuffer.size(), AudBufSize,
                (unsigned)VideoBuffer.size(), VidBufSize,
                A_sent, V_sent);
            /**/
        }
    }
    std::string GetMEncoderRawvideoParam() const
    {
        char Buf[512];
        unsigned bpp   = INPUT_BPP; if(bpp == 15 || bpp == 17) bpp = 16;
        sprintf(Buf, "fps=%.8g:format=%s:w=%u:h=%u:size=%u",
            vid_fps_scaled / (double)FPS_SCALE,
            USE_IMGFMT,
            vid_width,
            vid_height,
            vid_width*vid_height * bpp/8);
        return Buf;
    }
    std::string GetMEncoderRawaudioParam() const
    {
        char Buf[512];
        sprintf(Buf, "channels=%u:rate=%u:samplesize=%u:bitrate=%u",
            aud_chans,
            aud_rate,
            aud_bits/8,
            aud_rate*aud_chans*(aud_bits/8) );
        return Buf;
    }
    std::string GetMEncoderCommand() const
    {
        std::string mandatory = "-audiofile " + AUDIO_FN
                              + " -audio-demuxer rawaudio"
                              + " -demuxer rawvideo"
                              + " -rawvideo " + GetMEncoderRawvideoParam()
                              + " -rawaudio " + GetMEncoderRawaudioParam()
                              ;
        std::string cmd = VIDEO_CMD;

        std::string::size_type p = cmd.find("NESV""SETTINGS");
        if(p != cmd.npos)
            cmd = cmd.replace(p, 4+8, mandatory);
        else
            fprintf(stderr, "Warning: NESVSETTINGS not found in videocmd\n");

        char videonumstr[64];
        sprintf(videonumstr, "%u", videonumber);

        for(;;)
        {
            p = cmd.find("VIDEO""NUMBER");
            if(p == cmd.npos) break;
            cmd = cmd.replace(p, 5+6, videonumstr);
        }

        fprintf(stderr, "Launch: %s\n", cmd.c_str()); fflush(stderr);

        return cmd;
    }

    void CheckBegin()
    {
        if(!audfp)
        {
            unlink(AUDIO_FN.c_str());
            mknod(AUDIO_FN.c_str(), S_IFIFO|0666, 0);
        }

        if(!vidfp)
        {
            /* Note: popen does not accept b/t in mode param */
            setenv("LD_PRELOAD", "", 1);
            vidfp = popen(GetMEncoderCommand().c_str(), "w");
            if(!vidfp)
            {
                perror("Launch failed");
            }
            else
            {
                fcntl(fileno(vidfp), F_SETFL, O_WRONLY | O_NONBLOCK);
            }
        }

        if(!audfp)
        {
        Retry:
            audfp = fopen(AUDIO_FN.c_str(), "wb");

            if(!audfp)
            {
                perror(AUDIO_FN.c_str());
                if(errno == ESTALE) goto Retry;
            }
            else
            {
                fcntl(fileno(audfp), F_SETFL, O_WRONLY | O_NONBLOCK);
            }
        }
    }
};

static AVI* AVI = 0;

#include "quantize.hh"
#include "rgbtorgb.hh"

static long CurrentFrameNumber = 0;

extern "C"
{
    int LoggingEnabled = 0; /* 0=no, 1=yes, 2=recording! */

    const char* NESVideoGetVideoCmd()
    {
        return VIDEO_CMD.c_str();
    }
    void NESVideoSetVideoCmd(const char *cmd)
    {
        VIDEO_CMD = cmd;
    }

    static class AVI& GetAVIptr()
    {
        if(!AVI)
        {
            fprintf(stderr, "Starting new AVI (num %u)\n", videonumber);
            AVI = new NormalAVI;
        }
        return *AVI;
    }

    void NESVideoNextAVI()
    {
        if(AVI)
        {
            fprintf(stderr, "Closing AVI (next will be started)\n");
            delete AVI;
            AVI = 0;
            ++videonumber;
        }
    }

    static bool SubstituteWithBlackIfNeeded(const void*& data,
        unsigned width, unsigned height)
    {
        /* If the first frames of the animation consist of a
         * single color (such as gray for NES), replace them
         * with black to avoid ugly backgrounds on logo animations
         */

        static bool Deviate = false;
        static short* Replacement = 0;
        static unsigned wid=0, hei=0;
        if(Deviate)
        {
            if(Replacement) { delete[] Replacement; Replacement=0; }
            return false;
        }

        unsigned dim = width * height;
        const short* p = (const short*)data;
        for(unsigned a=0; a<dim; ++a)
            if(p[a] != p[0])
            {
                Deviate = true;
                return false;
            }

        if(Replacement && (wid != width || hei != height))
        {
            delete[] Replacement;
            Replacement = 0;
        }

        wid = width;
        hei = height;

        if(!Replacement)
        {
            Replacement = new short[dim];
            for(unsigned a=0; a<dim; ++a) Replacement[a]=0x0000;
        }
        data = (void*)Replacement;
        return true;
    }

    void NESVideoLoggingVideoPtr
        (const void*data, unsigned width,unsigned height,
         unsigned fps_scaled,
         unsigned bpp
        )
    {
        if(LoggingEnabled < 2) return;

        ++CurrentFrameNumber;

        if(bpp) INPUT_BPP = bpp;

        switch(INPUT_BPP)
        {
            case 32: USE_IMGFMT = "0x42475220" /*BGR32*/; break;
            case 24: USE_IMGFMT = "0x42475218" /*BGR24*/; break;
            case 16: USE_IMGFMT = "0x42475210" /*BGR16*/; break;
            case 15: USE_IMGFMT = "0x4247520F" /*BGR15*/; break;
            case 12: USE_IMGFMT = "0x30323449" /*I420*/; break;
            case 17: USE_IMGFMT = "0x32595559" /*YUY2*/; break;
            case 8: USE_IMGFMT = "rgb8"; break;
            default: USE_IMGFMT = 0;
        }

        if(INPUT_BPP == 16 || INPUT_BPP == 15)
        {
            static bool done_unicolor_frames = false;
            if(!done_unicolor_frames
            && !SubstituteWithBlackIfNeeded(data, width,height))
                done_unicolor_frames = true;
        }

        if(!USE_IMGFMT)
        {
            fprintf(stderr, "NESVIDEOS_PIECE does not know how to handle input bpp '%d'\n",
                INPUT_BPP);
            return;
        }

        GetAVIptr().VideoPtr(width,height,fps_scaled,  (const unsigned char*) data);
    }

    void NESVideoLoggingAudioPtr
        (const void*data,
         unsigned rate, unsigned bits, unsigned chans,
         unsigned nsamples)
    {
        if(LoggingEnabled < 2) return;

        ++CurrentFrameNumber;

        /*
        fprintf(stderr, "Writing %u samples (%u bits, %u chans, %u rate)\n",
            nsamples, bits, chans, rate);*/

        /*
        static FILE*fp = fopen("audiodump.wav", "wb");
        fwrite(data, 1, nsamples*(bits/8)*chans, fp);
        fflush(fp);*/

        GetAVIptr().AudioPtr(rate,bits,chans, (const unsigned char*) data, nsamples);
    }
} /* extern "C" */
