/*
 Ordered dithering methods provided for:
   8x8 (Quantize8x8)
   4x4 (Quantize4x4)
   3x3 (Quantize3x3)
   4x2 (Quantize4x2)
   3x2 (Quantize3x2)
   2x2 (Quantize2x2)
 The functions are:

   template<int m, int in_max>
   int QuantizeFunc(size_t quant_pos, double value)

      - Quantizes value, assumed to be in range 0..in_max, to range 0..m
      - quant_pos tells the coordinate into the dithering matrix

   template<int m, int in_max>
   int QuantizeFunc(size_t quant_pos, unsigned value)

      - Quantizes value, assumed to be in range 0..in_max, to range 0..m
      - quant_pos tells the coordinate into the dithering matrix

   For example:
     int o16 = (x + y*4) % 16;
     int c = Quantize4x4<0x1F,0xFF> (o16, value);
     - value, which is assumed to be in range 0..0xFF (inclusive),
       will be converted into range 0..0x1F (inclusive).
       o16 describes the dithering matrix position.

   If in_max is omitted, 0xFF is assumed for input value range.
   That is, Quantize4x4<v> is the same as Quantize4x4<v,0xFF>.

 Copyright (C) 2008,2010 Joel Yliluoma (http://iki.fi/bisqwit/)
*/

#define OrderedDitherDecl(n) \
    static const double flts[n]; \
    static const int ints[n]; \
    enum { mul = n+1, \
           maxin = in_max, \
           even = !(maxin % mul), \
           intmul = even ? 1 : mul };

/* macroes for initializing dither tables */
#define d(n) (n)/double(mul) - 0.5
#define i(n) even ? (n*in_max/mul - (int)in_max/2) \
                  : (n*in_max - (int)mul*in_max/2)

template<int m, int in_max = 255>
struct QuantizeNoDither
{
    int res;
    template<typename IntType>
    QuantizeNoDither(IntType v) : res(v * m / in_max) { }
    operator int() const { return res; }
};

template<int m, typename Base>
struct QuantizeFuncBase: private Base
{
    int res;

    QuantizeFuncBase(size_t quant_pos, double v) : res(0)
    {
        if(v > 0.0)
        {
            const double dither_threshold = Base::flts[quant_pos];
            res = (int)(v * (m / double(Base::maxin)) + dither_threshold);
            if(res > m) res = m;
        }
    }

    QuantizeFuncBase(size_t quant_pos, unsigned char v) : res(v)
    {
        if(m == Base::maxin) return;
        if(m < Base::maxin)
        {
            // With dithering
            const int dither_threshold = Base::ints[quant_pos];
            const int intmul = Base::intmul;
            res = (res * (m * intmul) + dither_threshold) / (Base::maxin * intmul);
        }
        else
        {
            // Without dithering
            res = QuantizeNoDither<m, Base::maxin> (res);
        }
    }
};

#define QuantizeFuncDecl(name, base) \
  template<int m, int in_max=255> \
  struct name: private QuantizeFuncBase<m, base<in_max> > \
  { \
      typedef QuantizeFuncBase<m, base<in_max> > Base; \
      template<typename A, typename B> name(A a, B b) : Base(a, b) { } \
      operator int() const { return Base::res; } \
  }

/******* Quantizing with 16x16 ordered dithering ********/
template<int in_max> struct OrderedDither_16x16 { OrderedDitherDecl(16*16) };
    template<int in_max>
    const double OrderedDither_16x16<in_max>::flts[] /* A table for 16x16 ordered dithering */
    = { d(  1), d(193), d( 49), d(241), d( 13), d(205), d( 61), d(253), d(  4), d(196), d( 52), d(244), d( 16), d(208), d( 64), d(256),
        d(129), d( 65), d(177), d(113), d(141), d( 77), d(189), d(125), d(132), d( 68), d(180), d(116), d(144), d( 80), d(192), d(128),
        d( 33), d(225), d( 17), d(209), d( 45), d(237), d( 29), d(221), d( 36), d(228), d( 20), d(212), d( 48), d(240), d( 32), d(224),
        d(161), d( 97), d(145), d( 81), d(173), d(109), d(157), d( 93), d(164), d(100), d(148), d( 84), d(176), d(112), d(160), d( 96),
        d(  9), d(201), d( 57), d(249), d(  5), d(197), d( 53), d(245), d( 12), d(204), d( 60), d(252), d(  8), d(200), d( 56), d(248),
        d(137), d( 73), d(185), d(121), d(133), d( 69), d(181), d(117), d(140), d( 76), d(188), d(124), d(136), d( 72), d(184), d(120),
        d( 41), d(233), d( 25), d(217), d( 37), d(229), d( 21), d(213), d( 44), d(236), d( 28), d(220), d( 40), d(232), d( 24), d(216),
        d(169), d(105), d(153), d( 89), d(165), d(101), d(149), d( 85), d(172), d(108), d(156), d( 92), d(168), d(104), d(152), d( 88),
        d(  3), d(195), d( 51), d(243), d( 15), d(207), d( 63), d(255), d(  2), d(194), d( 50), d(242), d( 14), d(206), d( 62), d(254),
        d(131), d( 67), d(179), d(115), d(143), d( 79), d(191), d(127), d(130), d( 66), d(178), d(114), d(142), d( 78), d(190), d(126),
        d( 35), d(227), d( 19), d(211), d( 47), d(239), d( 31), d(223), d( 34), d(226), d( 18), d(210), d( 46), d(238), d( 30), d(222),
        d(163), d( 99), d(147), d( 83), d(175), d(111), d(159), d( 95), d(162), d( 98), d(146), d( 82), d(174), d(110), d(158), d( 94),
        d( 11), d(203), d( 59), d(251), d(  7), d(199), d( 55), d(247), d( 10), d(202), d( 58), d(250), d(  6), d(198), d( 54), d(246),
        d(139), d( 75), d(187), d(123), d(135), d( 71), d(183), d(119), d(138), d( 74), d(186), d(122), d(134), d( 70), d(182), d(118),
        d( 43), d(235), d( 27), d(219), d( 39), d(231), d( 23), d(215), d( 42), d(234), d( 26), d(218), d( 38), d(230), d( 22), d(214),
        d(171), d(107), d(155), d( 91), d(167), d(103), d(151), d( 87), d(170), d(106), d(154), d( 90), d(166), d(102), d(150), d( 86) };
    template<int in_max>
    const int OrderedDither_16x16<in_max>::ints[]
    = { i(  1), i(193), i( 49), i(241), i( 13), i(205), i( 61), i(253), i(  4), i(196), i( 52), i(244), i( 16), i(208), i( 64), i(256),
        i(129), i( 65), i(177), i(113), i(141), i( 77), i(189), i(125), i(132), i( 68), i(180), i(116), i(144), i( 80), i(192), i(128),
        i( 33), i(225), i( 17), i(209), i( 45), i(237), i( 29), i(221), i( 36), i(228), i( 20), i(212), i( 48), i(240), i( 32), i(224),
        i(161), i( 97), i(145), i( 81), i(173), i(109), i(157), i( 93), i(164), i(100), i(148), i( 84), i(176), i(112), i(160), i( 96),
        i(  9), i(201), i( 57), i(249), i(  5), i(197), i( 53), i(245), i( 12), i(204), i( 60), i(252), i(  8), i(200), i( 56), i(248),
        i(137), i( 73), i(185), i(121), i(133), i( 69), i(181), i(117), i(140), i( 76), i(188), i(124), i(136), i( 72), i(184), i(120),
        i( 41), i(233), i( 25), i(217), i( 37), i(229), i( 21), i(213), i( 44), i(236), i( 28), i(220), i( 40), i(232), i( 24), i(216),
        i(169), i(105), i(153), i( 89), i(165), i(101), i(149), i( 85), i(172), i(108), i(156), i( 92), i(168), i(104), i(152), i( 88),
        i(  3), i(195), i( 51), i(243), i( 15), i(207), i( 63), i(255), i(  2), i(194), i( 50), i(242), i( 14), i(206), i( 62), i(254),
        i(131), i( 67), i(179), i(115), i(143), i( 79), i(191), i(127), i(130), i( 66), i(178), i(114), i(142), i( 78), i(190), i(126),
        i( 35), i(227), i( 19), i(211), i( 47), i(239), i( 31), i(223), i( 34), i(226), i( 18), i(210), i( 46), i(238), i( 30), i(222),
        i(163), i( 99), i(147), i( 83), i(175), i(111), i(159), i( 95), i(162), i( 98), i(146), i( 82), i(174), i(110), i(158), i( 94),
        i( 11), i(203), i( 59), i(251), i(  7), i(199), i( 55), i(247), i( 10), i(202), i( 58), i(250), i(  6), i(198), i( 54), i(246),
        i(139), i( 75), i(187), i(123), i(135), i( 71), i(183), i(119), i(138), i( 74), i(186), i(122), i(134), i( 70), i(182), i(118),
        i( 43), i(235), i( 27), i(219), i( 39), i(231), i( 23), i(215), i( 42), i(234), i( 26), i(218), i( 38), i(230), i( 22), i(214),
        i(171), i(107), i(155), i( 91), i(167), i(103), i(151), i( 87), i(170), i(106), i(154), i( 90), i(166), i(102), i(150), i( 86) };
QuantizeFuncDecl(Quantize16x16, OrderedDither_16x16);


/******* Quantizing with 8x8 ordered dithering ********/
template<int in_max> struct OrderedDither_8x8 { OrderedDitherDecl(8*8) };
    template<int in_max>
    const double OrderedDither_8x8<in_max>::flts[] /* A table for 8x8 ordered dithering */
    = { d(1 ), d(49), d(13), d(61), d( 4), d(52), d(16), d(64),
        d(33), d(17), d(45), d(29), d(36), d(20), d(48), d(32),
        d(9 ), d(57), d( 5), d(53), d(12), d(60), d( 8), d(56),
        d(41), d(25), d(37), d(21), d(44), d(28), d(40), d(24),
        d(3 ), d(51), d(15), d(63), d( 2), d(50), d(14), d(62),
        d(35), d(19), d(47), d(31), d(34), d(18), d(46), d(30),
        d(11), d(59), d( 7), d(55), d(10), d(58), d( 6), d(54),
        d(43), d(27), d(39), d(23), d(42), d(26), d(38), d(22) };
    template<int in_max>
    const int OrderedDither_8x8<in_max>::ints[]
    = { i(1 ), i(49), i(13), i(61), i( 4), i(52), i(16), i(64),
        i(33), i(17), i(45), i(29), i(36), i(20), i(48), i(32),
        i(9 ), i(57), i( 5), i(53), i(12), i(60), i( 8), i(56),
        i(41), i(25), i(37), i(21), i(44), i(28), i(40), i(24),
        i(3 ), i(51), i(15), i(63), i( 2), i(50), i(14), i(62),
        i(35), i(19), i(47), i(31), i(34), i(18), i(46), i(30),
        i(11), i(59), i( 7), i(55), i(10), i(58), i( 6), i(54),
        i(43), i(27), i(39), i(23), i(42), i(26), i(38), i(22) };
QuantizeFuncDecl(Quantize8x8, OrderedDither_8x8);


/******* Quantizing with 4x4 ordered dithering ********/
template<int in_max> struct OrderedDither_4x4 { OrderedDitherDecl(4*4) };
    template<int in_max>
    const double OrderedDither_4x4<in_max>::flts[] /* A table for 4x4 ordered dithering */
    = { d( 1), d( 9), d( 3), d(11),
        d(13), d( 5), d(15), d( 7),
        d( 4), d(12), d( 2), d(10),
        d(16), d( 8), d(14), d( 6) };
    template<int in_max>
    const int OrderedDither_4x4<in_max>::ints[]
    = { i( 1), i( 9), i( 3), i(11),
        i(13), i( 5), i(15), i( 7),
        i( 4), i(12), i( 2), i(10),
        i(16), i( 8), i(14), i( 6) };
QuantizeFuncDecl(Quantize4x4, OrderedDither_4x4);

/******* Quantizing with 3x3 ordered dithering (DEPRECATED) ********/
template<int in_max> struct OrderedDither_3x3 { OrderedDitherDecl(3*3) };
    template<int in_max>
    const double OrderedDither_3x3<in_max>::flts[] /* A table for 3x3 ordered dithering */
    = { d(1), d(7), d(3),
        d(6), d(4), d(9),
        d(8), d(2), d(5) };
    template<int in_max>
    const int OrderedDither_3x3<in_max>::ints[]
    = { i(1), i(7), i(3),
        i(6), i(4), i(9),
        i(8), i(2), i(5) };
QuantizeFuncDecl(Quantize3x3, OrderedDither_3x3);

/******* Quantizing with 4x2 ordered dithering ********/
template<int in_max> struct OrderedDither_4x2 { OrderedDitherDecl(4*2) };
    template<int in_max>
    const double OrderedDither_4x2<in_max>::flts[] /* A table for 4x2 ordered dithering */
    = { d(1), d(5), d(2), d(6),
        d(7), d(3), d(8), d(4) };
    template<int in_max>
    const int OrderedDither_4x2<in_max>::ints[]
    = { i(1), i(5), i(2), i(6),
        i(7), i(3), i(8), i(4) };
QuantizeFuncDecl(Quantize4x2, OrderedDither_4x2);

/******* Quantizing with 3x2 ordered dithering (DEPRECATED) ********/
template<int in_max> struct OrderedDither_3x2 { OrderedDitherDecl(3*2) };
    template<int in_max>
    const double OrderedDither_3x2<in_max>::flts[] /* A table for 3x2 ordered dithering */
    = { d(1), d(5), d(3),
        d(4), d(2), d(6) };
    template<int in_max>
    const int OrderedDither_3x2<in_max>::ints[]
    = { i(1), i(5), i(3),
        i(4), i(2), i(6) };
QuantizeFuncDecl(Quantize3x2, OrderedDither_3x2);

/******* Quantizing with 2x2 ordered dithering ********/
template<int in_max> struct OrderedDither_2x2 { OrderedDitherDecl(2*2) };
    template<int in_max>
    const double OrderedDither_2x2<in_max>::flts[] /* A table for 2x2 ordered dithering */
    = { d(1), d(4),
        d(3), d(2) };
    template<int in_max>
    const int OrderedDither_2x2<in_max>::ints[]
    = { i(1), i(4),
        i(3), i(2) };
QuantizeFuncDecl(Quantize2x2, OrderedDither_2x2);


#undef OrderedDitherDecl
#undef QuantizeFuncDecl
#undef i
#undef d
